import sys
import os

from ..helpers.extended_json import ejson

ACTIVE_SUFFIX = '.active'


def stream(file=sys.stdout):

    filename = None

    if isinstance(file, str):
        filename = file + ACTIVE_SUFFIX
        basedir = os.path.dirname(filename)
        os.makedirs(basedir, exist_ok=True)
        file = open(filename, 'w')

    def write(obj):
        file.write(ejson.dumps(obj, sort_keys=True, ensure_ascii=True)+'\n')
        file.flush()

    def res_writer(res):
        for r in res:
            write(r)
            yield r

    def func(package):
        write(package.pkg.descriptor)
        yield package.pkg
        for res in package:
            yield res_writer(res)
            file.write('\n')
        file.close()
        if filename:
            os.rename(filename, filename[:-len(ACTIVE_SUFFIX)])

    return func



from inspect import signature

from .. import DataStreamProcessor


class finalizer(DataStreamProcessor):

    def __init__(self, callback):
        super().__init__()
        self.callback = callback

    def get_iterator(self, datastream):
        base_func = super().get_iterator(datastream)

        def func():
            yield from base_func()
            if 'stats' in signature(self.callback).parameters:
                stats = datastream.merge_stats()
                self.callback(stats=stats)
            else:
                self.callback()
        return func



import itertools
import decimal
import datetime

from datapackage import Package, Resource
from tableschema.storage import Storage

from .. import DataStreamProcessor


class iterable_storage(Storage):

    SAMPLE_SIZE = 100

    def __init__(self, iterable):
        super(iterable_storage, self).__init__()
        self.iterable = iterable
        self.schema = None

    def connect(self, name): pass
    def buckets(self): pass
    def create(self): pass
    def delete(self): pass
    def read(self): pass
    def write(self): pass

    def field_type(self, values):
        types = set()
        for value in values:
            if isinstance(value, str):
                types.add('string')
            elif isinstance(value, bool):
                types.add('boolean')
            elif isinstance(value, int):
                types.add('integer')
            elif isinstance(value, (float, decimal.Decimal)):
                types.add('number')
            elif isinstance(value, list):
                types.add('array')
            elif isinstance(value, dict):
                types.add('object')
            elif isinstance(value, datetime.datetime):
                types.add('datetime')
            elif isinstance(value, datetime.date):
                types.add('date')
            elif value is None:
                pass
            else:
                assert 'Unknown Python type: %r' % value
        if len(types) != 1:
            return 'any'
        else:
            return types.pop()

    def describe(self, _, descriptor=None):
        if descriptor is not None:
            return descriptor
        if self.schema is None:
            try:
                sample = list(itertools.islice(self.iterable, self.SAMPLE_SIZE))
                rec = sample[0]
                self.iterable = itertools.chain(sample, self.iterable)
                self.schema = dict(
                    fields=[
                        dict(name=name,
                             type=self.field_type([s.get(name) for s in sample]))
                        for name in rec.keys()
                    ]
                )
            except Exception:
                self.schema = dict(fields=[])
        return self.schema

    def iter(self, _):
        return self.iterable


class iterable_loader(DataStreamProcessor):

    def __init__(self, iterable, name=None):
        super(iterable_loader, self).__init__()
        self.iterable = iterable
        self.name = name
        self.exc = None

    def handle_iterable(self):
        mode = None
        try:
            for x in self.iterable:
                if mode is None:
                    assert isinstance(x, (dict, list, tuple)), 'Bad item %r' % x
                    mode = dict if isinstance(x, dict) else list
                assert isinstance(x, mode)
                if mode == dict:
                    yield x
                else:
                    yield dict(zip(('col{}'.format(i) for i in range(len(x))), x))
        except Exception as e:
            self.exc = e
            raise

    def process_datapackage(self, dp: Package):
        name = self.name
        if name is None:
            name = 'res_{}'.format(len(dp.resources) + 1)
        self.res = Resource(dict(
            name=name,
            path='{}.csv'.format(name)
        ), storage=iterable_storage(self.handle_iterable()))
        self.res.infer()
        if self.exc is not None:
            raise self.exc
        dp.descriptor.setdefault('resources', []).append(self.res.descriptor)
        return dp

    def process_resources(self, resources):
        yield from super(iterable_loader, self).process_resources(resources)
        yield self.res.iter(keyed=True)



from .. import DataStreamProcessor


class row_processor(DataStreamProcessor):

    def __init__(self, row_processor_func):
        super(row_processor, self).__init__()
        self.func = row_processor_func

    def process_row(self, row):
        ret = self.func(row)
        if ret is None:
            return row
        return ret



from dataflows import PackageWrapper
from dataflows.helpers.resource_matcher import ResourceMatcher


def set_primary_key(primary_key, resources=None):

    def func(package: PackageWrapper):
        matcher = ResourceMatcher(resources, package.pkg)
        for resource in package.pkg.descriptor['resources']:
            if matcher.match(resource['name']):
                resource.setdefault('schema', {})['primaryKey'] = primary_key
        yield package.pkg

        res_iter = iter(package)
        for r in res_iter:
            if matcher.match(r.res.name):
                yield r.it
            else:
                yield r

    return func



from dataflows import PackageWrapper, ResourceWrapper

from ..helpers.resource_matcher import ResourceMatcher


def deduper(rows: ResourceWrapper):
    pk = rows.res.descriptor['schema'].get('primaryKey', [])
    if len(pk) == 0:
        yield from rows
    else:
        keys = set()
        for row in rows:
            key = tuple(row[k] for k in pk)
            if key in keys:
                continue
            keys.add(key)
            yield row


def deduplicate(resources=None):

    def func(package: PackageWrapper):
        resource_matcher = ResourceMatcher(resources, package)
        yield package.pkg
        resource: ResourceWrapper
        for resource in package:
            if resource_matcher.match(resource.res.name):
                yield deduper(resource)
            else:
                yield resource
    return func



from datapackage import Package


class DataStream:
    def __init__(self, dp=None, res_iter=None, stats=None):
        self.dp = dp if dp is not None else Package()
        self.res_iter = res_iter if res_iter is not None else []
        self.stats = stats if stats is not None else []

    def merge_stats(self):
        ret = {}
        for s in self.stats:
            ret.update(s)
        return ret

    def _process(self):
        return self



from . import exceptions
from .datastream import DataStream
from .datastream_processor import DataStreamProcessor
from .resource_wrapper import ResourceWrapper
from .package_wrapper import PackageWrapper
from .flow import Flow
from .schema_validator import schema_validator, ValidationError



import re


class ResourceMatcher(object):

    def __init__(self, resources, datapackage):
        self.resources = resources
        if resources is None:
            self.resources = None
        elif isinstance(self.resources, str):
            self.resources = re.compile('^' + self.resources + '$')
            self.re = True
        elif isinstance(self.resources, int):
            if isinstance(datapackage, dict):
                self.resources = [datapackage['resources'][self.resources]['name']]
            else:
                self.resources = [datapackage.resources[self.resources].name]
            self.re = False
        else:
            assert isinstance(self.resources, list)
            self.re = False

    def match(self, name):
        if self.resources is None:
            return True
        if self.re:
            return self.resources.match(name) is not None
        else:
            return name in self.resources



import logging
import itertools
import collections
import copy

from datapackage import Package
from tableschema.exceptions import CastError, UniqueKeyError

from . import exceptions
from .datastream import DataStream
from .resource_wrapper import ResourceWrapper
from .schema_validator import schema_validator, raise_exception


class LazyIterator:

    def __init__(self, get_iterator):
        self.get_iterator = get_iterator

    def __iter__(self):
        return self.get_iterator()


class DataStreamProcessor:

    def __init__(self):
        self.stats = {}
        self.source = None
        self.datapackage = None
        self.position = None

    def __call__(self, source=None, position=None):
        if source is None:
            source = DataStream()
        self.source = source
        self.position = position
        return self

    def process_resource(self, resource: ResourceWrapper):
        for row in resource:
            yield self.process_row(row)

    def process_resources(self, resources):
        for res in resources:
            yield self.process_resource(res)

    def process_row(self, row):
        return row

    def process_datapackage(self, dp: Package):
        return dp

    def get_res(self, current_dp, name):
        ret = self.datapackage.get_resource(name)
        if ret is None:
            ret = current_dp.get_resource(name)
        assert ret is not None
        return ret

    def get_iterator(self, datastream):
        current_dp = datastream.dp
        res_iter_ = datastream.res_iter

        def func():
            res_iter = (ResourceWrapper(self.get_res(current_dp, rw.res.name), rw.it)
                        for rw in res_iter_)
            res_iter = self.process_resources(res_iter)
            res_iter = (it if isinstance(it, ResourceWrapper) else ResourceWrapper(res, it)
                        for res, it
                        in itertools.zip_longest(self.datapackage.resources, res_iter))
            return res_iter
        return func

    def _process(self):
        datastream = self.source._process()

        try:
            self.datapackage = Package(descriptor=copy.deepcopy(datastream.dp.descriptor))
            self.datapackage = self.process_datapackage(self.datapackage)
            self.datapackage.commit()

            return DataStream(self.datapackage,
                            LazyIterator(self.get_iterator(datastream)),
                            datastream.stats + [self.stats])
        except Exception as exception:
            self.raise_exception(exception)

    def raise_exception(self, cause):
        if not isinstance(cause, exceptions.ProcessorError):
            error = exceptions.ProcessorError(
                cause,
                processor_name=self.__class__.__name__,
                processor_object=self,
                processor_position=self.position
            )
            raise error from cause
        raise cause

    def safe_process(self, on_error=None):
        results = []
        try:
            ds = self._process()
            for res in ds.res_iter:
                if on_error is not None:
                    results.append(list(
                        schema_validator(res.res, res, on_error=on_error)
                    ))
                else:
                    collections.deque(res, maxlen=0)
        except UniqueKeyError as e:
            self.raise_exception(e)
        except CastError as e:
            for err in e.errors:
                logging.error('%s', err)
        except Exception as exception:
            self.raise_exception(exception)
        return ds, results

    def process(self):
        ds, _ = self.safe_process()
        return ds.dp, ds.merge_stats()

    def results(self, on_error=None):
        if on_error is None:
            on_error = raise_exception
        ds, results = self.safe_process(on_error=on_error)
        return results, ds.dp, ds.merge_stats()



import re

from ..helpers.resource_matcher import ResourceMatcher


def _find_replace(rows, fields):
    for row in rows:
        for field in fields:
            for pattern in field.get('patterns', []):
                row[field['name']] = re.sub(
                    str(pattern['find']),
                    str(pattern['replace']),
                    str(row[field['name']]))
        yield row


def find_replace(fields, resources=None):

    def func(package):
        matcher = ResourceMatcher(resources, package.pkg)
        yield package.pkg
        for rows in package:
            if matcher.match(rows.res.name):
                yield _find_replace(rows, fields)
            else:
                yield rows

    return func



from .load import load
from .printer import printer
from .set_type import set_type
from .validate import validate
from .dumpers import dump_to_path, dump_to_zip, dump_to_sql

from .add_computed_field import add_computed_field
from .add_field import add_field
from .checkpoint import checkpoint
from .concatenate import concatenate
from .conditional import conditional
from .delete_fields import delete_fields
from .deduplicate import deduplicate
from .duplicate import duplicate
from .filter_rows import filter_rows
from .finalizer import finalizer
from .find_replace import find_replace
from .join import join, join_self, join_with_self
from .parallelize import parallelize
from .rename_fields import rename_fields
from .select_fields import select_fields
from .set_primary_key import set_primary_key
from .sort_rows import sort_rows
from .sources import sources
from .stream import stream
from .unpivot import unpivot
from .unstream import unstream
from .update_package import update_package, add_metadata
from .update_resource import update_resource
from .update_schema import update_schema
from .update_stats import update_stats



import os
import sys
import subprocess

import slugify
import click
from jinja2 import Environment, PackageLoader
import inquirer
from inquirer import themes

# Some settings
FORMATS = ['csv', 'tsv', 'xls', 'xlsx', 'json', 'ndjson', 'ods', 'gsheet']
INPUTS = dict((
  ('File', 'file'),
  ('Remote URL', 'remote'),
  ('SQL Database', 'sql'),
  ('Other', 'other')
))
PROCESSING = dict((
  ('Sort all rows by key',                'sort'),
  ('Filter according to column values',   'filter'),
  ('Search & replace values in the data', 'find_replace'),
  ('Delete some columns',                 'delete_fields'),
  ('Normalize and validate numbers, dates and other types', 'set_type'),
  ('Un-pivot the data',                   'unpivot'),
  ('Custom row-by-row processing',        'custom'),
))
OUTPUTS = dict((
  ('Just print the data',                     'print'),
  ('As a Python list',                        'list'),
  ('A CSV file (in a data package)',          'dp_csv'),
  ('A CSV file (in a zipped data package)',   'dp_csv_zip'),
  ('A JSON file (in a data package)',         'dp_json'),
  ('A JSON file (in a zipped data package)',  'dp_json_pkg'),
  ('An SQL database',                         'sql'),
))


# Utility functions
def fall(*validators):
    def func(*args):
        return all(v(*args) for v in validators)
    return func


def fany(*validators):
    def func(*args):
        return any(v(*args) for v in validators)
    return func


# Validators
def valid_url(ctx, url):
    return url.startswith('http://') or url.startswith('https://')


def not_empty(ctx, x):
    return x


# Converters
def convert_processing(ctx, key):
    ctx['processing'] = [PROCESSING[x] for x in key]
    return True


def convert_input(ctx, key):
    ctx['input'] = INPUTS[key]
    return True


def convert_output(ctx, key):
    ctx['output'] = OUTPUTS[key]
    return True


def extract_format(ctx, url):
    if url:
        _, ext = os.path.splitext(url)
        if ext:
            ctx['format'] = ext[1:].lower()
            return True
    ctx['format'] = None
    return True


# Render
env = Environment(loader=PackageLoader('dataflows'),
                  trim_blocks=True, lstrip_blocks=True,
                  autoescape=False)


def render(parameters):
    tpl = env.get_template('main.tpl.py')
    return tpl.render(**parameters)


@click.group()
def cli():
    pass


# Main CLI routine
@cli.command()
@click.argument('arg', default='interactive')
def init(arg):
    """Bootstrap a processing pipeline script.
ARG is either a path or a URL for some data to read from, 'hello-world' for a full working code example,
or leave empty for an interactive walkthrough.
    """

    answers = {'a': 1}
    if arg == 'interactive':
        input("""Hi There!
    DataFlows will now bootstrap a data processing flow based on your needs.

    Press any key to start...
    """)

    elif arg == 'hello-world':
        raise NotImplementedError()
    else:
        url = arg
        answers = dict(
            input='remote',
            title=os.path.basename(url),
            input_url=url,
            processing=[],
            output='print_n_pkg'
        )
        extract_format(answers, url)

    questions = [
        # Input
        inquirer.List('input_str',
                    message='What is the source of your data?',
                    choices=INPUTS.keys(),
                    ignore=lambda ctx: ctx.get('input') is not None,
                    validate=convert_input),

        # Input Parameters
        inquirer.Text('input_url',
                    message='What is the path of that file',
                    ignore=fany(lambda ctx: ctx.get('input') != 'file',
                                lambda ctx: ctx.get('input_url') is not None),
                    validate=fall(not_empty, extract_format)),
        inquirer.List('format',
                    message='We couldn''t detect the file format - which is it?',
                    choices=FORMATS[:-1],
                    ignore=fany(lambda ctx: ctx.get('input') != 'file',
                                lambda ctx: ctx.get('format') in FORMATS)),

        inquirer.Text('input_url',
                    message='Where is that file located (URL)',
                    ignore=fany(lambda ctx: ctx.get('input') != 'remote',
                                lambda ctx: ctx.get('input_url') is not None),
                    validate=fall(extract_format, not_empty, valid_url)),
        inquirer.List('format',
                      message='We couldn''t detect the source format - which is it',
                      choices=FORMATS,
                      ignore=fany(lambda ctx: ctx['input'] != 'remote',
                                  lambda ctx: ctx.get('format') in FORMATS)
                      ),
        inquirer.Text('sheet',
                      message='Which sheet in the spreadsheet should be processed (name or index)',
                      validate=not_empty,
                      ignore=lambda ctx: ctx.get('format') not in ('xls', 'xlsx', 'ods'),
                      ),
        inquirer.Text('input_url',
                      message='What is the connection string to the database',
                      validate=not_empty,
                      ignore=fany(lambda ctx: ctx['input'] != 'sql',
                                  lambda ctx: ctx.get('input_url') is not None),
                      ),
        inquirer.Text('input_db_table',
                      message='...and the name of the database table to extract',
                      validate=not_empty,
                      ignore=fany(lambda ctx: ctx['input'] != 'sql',
                                  lambda ctx: ctx.get('input_db_table') is not None),
                      ),

        inquirer.Text('input_url',
                      message='Describe that other source (shortly)',
                      ignore=fany(lambda ctx: ctx['input'] != 'other',
                                  lambda ctx: ctx.get('input_url') is not None),
                      ),

        # Processing
        inquirer.Checkbox('processing_str',
                        message='What kind of processing would you like to run on the data',
                        choices=PROCESSING.keys(),
                        ignore=lambda ctx: ctx.get('processing') is not None,
                        validate=convert_processing),

        # Output
        inquirer.List('output_str',
                    message='Finally, where would you like the output data',
                    choices=OUTPUTS.keys(),
                    ignore=lambda ctx: ctx.get('output') is not None,
                    validate=convert_output),
        inquirer.Text('output_url',
                      message='What is the connection string to the database',
                      validate=not_empty,
                      ignore=fany(lambda ctx: ctx['output'] != 'sql',
                                  lambda ctx: ctx.get('output_url') is not None),
                      ),
        inquirer.Text('output_db_table',
                      message='...and the name of the database table to write to',
                      validate=not_empty,
                      ignore=fany(lambda ctx: ctx['output'] != 'sql',
                                  lambda ctx: ctx.get('output_db_table') is not None),
                      ),

        # # Finalize
        inquirer.Text('title',
                    message='That''s it! Now, just provide a title for your processing flow',
                    ignore=lambda ctx: ctx.get('title') is not None,
                    validate=not_empty),
    ]
    answers = inquirer.prompt(questions, answers=answers, theme=themes.GreenPassion())
    if answers is None:
        return
    answers['slug'] = slugify.slugify(answers['title'], separator='_')

    filename = '{slug}.py'.format(**answers)
    with open(filename, 'w') as out:
        print('Writing processing code into {}'.format(filename))
        out.write(render(answers))

    try:
        print('Running {}'.format(filename))
        ret = subprocess.check_output(sys.executable + ' ' + filename,
                                      stderr=subprocess.PIPE, universal_newlines=True, shell=True)
        print(ret)
        print('Done!')
    except subprocess.CalledProcessError as e:
        print('Processing failed, here''s the error:')
        print(e.stderr)
        answers = inquirer.prompt([
            inquirer.Confirm('edit',
                            message='Would you like to open {} in the default editor?'.format(filename),
                            default=False)
        ])
        if answers['edit']:
            click.edit(filename=filename)


if __name__ == '__main__':
    sys.exit(cli())



import os
import shutil

from .file_dumper import FileDumper


class PathDumper(FileDumper):

    def __init__(self, out_path='.', **options):
        super(PathDumper, self).__init__(options)
        self.out_path = out_path
        PathDumper.__makedirs(self.out_path)

    def write_file_to_output(self, filename, path):
        path = os.path.join(self.out_path, path)
        # Avoid rewriting existing files
        if self.add_filehash_to_path and os.path.exists(path):
            return
        path_part = os.path.dirname(path)
        PathDumper.__makedirs(path_part)
        shutil.copy(filename, path)
        return path

    @staticmethod
    def __makedirs(path):
        os.makedirs(path, exist_ok=True)



import re

from ..helpers.resource_matcher import ResourceMatcher


def process_resource(rows, fields):
    for row in rows:
        yield dict(
            (fields.get(k, k), v)
            for k, v in row.items()
        )


def rename_fields(fields, resources=None, regex=True):

    def func(package):
        matcher = ResourceMatcher(resources, package.pkg)
        dp_resources = package.pkg.descriptor.get('resources', [])
        field_res = [
            (re.compile(
                '^{}$'.format(
                    src if regex else re.escape(src)
                )
            ), tgt) for src, tgt in fields.items()
        ]
        matched = set()
        renames = dict()
        renamed_fields = dict()
        for resource in dp_resources:
            res_name = resource['name']
            renamed_fields.setdefault(res_name, set())
            renames.setdefault(res_name, dict())
            if matcher.match(res_name):
                schema_fields = resource['schema'].get('fields', [])
                for sf in schema_fields:
                    sf_name = sf['name']
                    for src, tgt in field_res:
                        if src.match(sf_name):
                            matched.add(src.pattern)
                            target_name = src.sub(tgt, sf_name)
                            assert target_name not in renamed_fields[res_name],\
                                f'Renaming two fields to the same name "{target_name}"'
                            renamed_fields[res_name].add(target_name)
                            renames[res_name][sf_name] = target_name
                            sf['name'] = target_name
                            break
        not_matched = [
            src.pattern for src, _ in field_res
            if src.pattern not in matched
        ]
        if len(not_matched) > 0:
            print('WARNING: Failed to match these fields to rename {!r}'.format(not_matched))
        yield package.pkg

        for resource in package:
            if not matcher.match(resource.res.name):
                yield resource
            else:
                yield process_resource(resource, renames[resource.res.name])

    return func



from inspect import signature

from datapackage import Resource
from tableschema import Schema
from tableschema.exceptions import CastError


class ValidationError(Exception):

    def __init__(self, resource_name, row, index, cast_error):
        msg = '\nROW: %r\n----\n' % row
        if cast_error is not None and hasattr(cast_error, 'errors'):
            msg += '\n'.join('%d) %s' % (i+1, err)
                            for i, err
                            in enumerate(cast_error.errors))
        super().__init__(msg)
        self.resource_name = resource_name
        self.row = row
        self.index = index
        self.cast_error = cast_error


def raise_exception(res_name, row, i, e):
    raise ValidationError(res_name, row, i, e)


def ignore(res_name, row, i, e):
    return True


def drop(res_name, row, i, e):
    return False


def clear(res_name, row, i, e, field):
    if field is not None:
        row[field.name] = None
        return True
    else:
        return False


def wrap_handler(on_error):
    assert callable(on_error)
    if len(list(signature(on_error).parameters)) > 4:
        return on_error

    def func(res_name, row, i, e, _):
        return on_error(res_name, row, i, e)
    return func


def schema_validator(resource, iterator,
                     field_names=None, on_error=None):
    if on_error is None:
        on_error = raise_exception
    on_error = wrap_handler(on_error)

    if isinstance(resource, Resource):
        schema: Schema = resource.schema
        assert schema is not None
        resource = resource.descriptor
    else:
        schema: Schema = Schema(resource.get('schema', {}))
    if field_names is None:
        field_names = [f.name for f in schema.fields]
    schema_fields = [f for f in schema.fields if f.name in field_names]
    for i, row in enumerate(iterator):
        field = None
        okay = True
        for field in schema_fields:
            try:
                row[field.name] = field.cast_value(row.get(field.name))
            except CastError as e:
                if not on_error(resource['name'], row, i, e, field):
                    okay = False
        if okay:
            yield row


schema_validator.drop = drop
schema_validator.ignore = ignore
schema_validator.clear = clear
schema_validator.raise_exception = raise_exception



from tabulator.parser import Parser
from tabulator.helpers import reset_stream


class ExcelXMLParser(Parser):
    options = []

    def __init__(self, loader, force_parse, **options):
        self.__loader = loader
        self.__force_parse = force_parse
        self.__extended_rows = None
        self.__encoding = None
        self.__chars = None

    def open(self, source, encoding=None):
        self.close()
        self.__chars = self.__loader.load(source, encoding=encoding)
        self.__encoding = getattr(self.__chars, 'encoding', encoding)
        if self.__encoding:
            self.__encoding.lower()
        self.reset()

    def close(self):
        if not self.closed:
            self.__chars.close()

    def reset(self):
        reset_stream(self.__chars)
        self.__extended_rows = self.__iter_extended_rows()

    @property
    def closed(self):
        return self.__chars is None

    @property
    def encoding(self):
        return self.__encoding

    @property
    def extended_rows(self):
        return self.__extended_rows

    # Private

    def __iter_extended_rows(self):
        from xml.sax import ContentHandler, parse

        class ExcelHandler(ContentHandler):
            def __init__(self):
                self.chars = []
                self.cells = []
                self.rows = []
                self.tables = []

            def characters(self, content):
                self.chars.append(content)

            def startElement(self, name, atts):
                if name == 'Cell':
                    self.chars = []
                elif name == 'Row':
                    self.cells = []
                elif name == 'Table':
                    self.rows = []

            def endElement(self, name):
                if name == 'Cell':
                    self.cells.append(''.join(self.chars))
                elif name == 'Row':
                    self.rows.append(self.cells)
                elif name == 'Table':
                    self.tables.append(self.rows)

        excelHandler = ExcelHandler()
        parse(self.__chars, excelHandler)
        headers = excelHandler.tables[0][0]
        rows = excelHandler.tables[0][1:]
        for row_number, row in enumerate(rows, start=1):
            yield (row_number, headers, row)



import copy
import re

from ..helpers.resource_matcher import ResourceMatcher


def match_fields(field_name_re, expected):
    def _filter(field):
        return (field_name_re.fullmatch(field['name']) is not None) is expected
    return _filter


def unpivot_rows(rows, fields_to_unpivot, fields_to_keep, extra_value):
    for row in rows:
        for unpivot_field in fields_to_unpivot:
            new_row = copy.deepcopy(unpivot_field['keys'])
            for field in fields_to_keep:
                new_row[field] = row[field]
            new_row[extra_value['name']] = row.get(unpivot_field['name'])
            yield new_row


def unpivot(unpivot_fields, extra_keys, extra_value, regex=True, resources=None):

    def func(package):

        matcher = ResourceMatcher(resources, package.pkg)
        all_res_config = {}
        for resource in package.pkg.descriptor['resources']:
            config = all_res_config.setdefault(resource['name'], {})
            name = resource['name']
            if not matcher.match(name):
                continue
            schema = resource.get('schema')
            if schema is None:
                continue

            fields = schema.get('fields', [])

            for u_field in unpivot_fields:
                if regex:
                    field_name_re = re.compile(u_field['name'])
                    fields_to_pivot = list(
                        filter(match_fields(field_name_re, True), fields)
                    )
                    fields = list(
                        filter(match_fields(field_name_re, False), fields)
                    )
                else:
                    field_name = u_field['name']
                    fields_to_pivot = list(
                        filter(lambda f: f['name'] == field_name, fields)
                    )
                    fields = list(
                        filter(lambda f: f['name'] != field_name, fields)
                    )

                # handle with regex
                config.setdefault('unpivot_fields_without_regex', [])
                for field_to_pivot in fields_to_pivot:
                    original_key_values = u_field['keys']  # With regex
                    new_key_values = {}
                    for key in original_key_values:
                        new_val = original_key_values[key]
                        if regex and isinstance(new_val, str):
                            new_val = re.sub(
                                u_field['name'], new_val, field_to_pivot['name'])
                        new_key_values[key] = new_val
                    field_to_pivot['keys'] = new_key_values
                    config['unpivot_fields_without_regex'].append(field_to_pivot)

            config['fields_to_keep'] = [f['name'] for f in fields]
            fields.extend(extra_keys)
            fields.append(extra_value)
            schema['fields'] = fields

        yield package.pkg

        for resource in package:
            if not matcher.match(resource.res.name):
                yield resource
            else:
                yield unpivot_rows(resource,
                                   all_res_config[resource.res.name]['unpivot_fields_without_regex'],
                                   all_res_config[resource.res.name]['fields_to_keep'],
                                   extra_value)

    return func



import re

from ..helpers.resource_matcher import ResourceMatcher


def process_resource(rows, fields):
    for row in rows:
        yield dict(
            (k, v)
            for k, v in row.items()
            if k in fields
        )


def delete_fields(fields, resources=None, regex=True):

    def func(package):
        matcher = ResourceMatcher(resources, package.pkg)
        dp_resources = package.pkg.descriptor.get('resources', [])
        field_res = [
            re.compile('^{}$'.format(f if regex else re.escape(f))) for f in fields
        ]
        matched = set()
        new_field_names = {}
        for resource in dp_resources:
            if matcher.match(resource['name']):
                schema_fields = resource['schema'].get('fields', [])
                new_fields = []
                for sf in schema_fields:
                    skip = False
                    for f in field_res:
                        if f.match(sf['name']):
                            skip = True
                            matched.add(f.pattern)
                            break
                    if not skip:
                        new_fields.append(sf)
                not_matched = [f for f in field_res if f.pattern not in matched]
                if len(not_matched) > 0:
                    print('WARNING: Failed to match these fields to delete {!r}'.format(not_matched))
                resource['schema']['fields'] = new_fields
                new_field_names[resource['name']] = [f['name'] for f in new_fields]
        yield package.pkg

        for resource in package:
            if not matcher.match(resource.res.name):
                yield resource
            else:
                yield process_resource(resource, new_field_names[resource.res.name])

    return func



import os
import hashlib
import json

from ... import DataStreamProcessor, ResourceWrapper, schema_validator


class DumperBase(DataStreamProcessor):

    def __init__(self, options={}):
        super(DumperBase, self).__init__()
        counters = options.get('counters', {})
        self.datapackage_rowcount = counters.get('datapackage-rowcount', 'count_of_rows')
        self.datapackage_bytes = counters.get('datapackage-bytes', 'bytes')
        self.datapackage_hash = counters.get('datapackage-hash', 'hash')
        self.resource_rowcount = counters.get('resource-rowcount', 'count_of_rows')
        self.resource_bytes = counters.get('resource-bytes', 'bytes')
        self.resource_hash = counters.get('resource-hash', 'hash')
        self.add_filehash_to_path = options.get('add_filehash_to_path', False)
        self.pretty_descriptor = options.get('pretty_descriptor', True)
        self.schema_validator_options = options.get('validator_options', {})

    @staticmethod
    def get_attr(obj, prop, default=None):
        if prop is None:
            return
        prop = prop.split('.')
        while len(prop) > 1:
            obj = obj.get(prop.pop(0), {})
        prop = prop.pop(0)
        return obj.get(prop, default)

    @staticmethod
    def set_attr(obj, prop, value):
        if prop is None:
            return
        prop = prop.split('.')
        while len(prop) > 1:
            obj = obj.setdefault(prop.pop(0), {})
        prop = prop.pop(0)
        obj[prop] = value

    @staticmethod
    def inc_attr(obj, prop, value):
        if prop is None:
            return
        prop = prop.split('.')
        while len(prop) > 1:
            obj = obj.setdefault(prop.pop(0), {})
        prop = prop.pop(0)
        obj.setdefault(prop, 0)
        obj[prop] += value

    @staticmethod
    def insert_hash_in_path(descriptor, hash):
        path = descriptor.get('path')
        if isinstance(path, list):
            if len(path) > 0:
                path = path[0]

        assert isinstance(path, str), '%r' % path

        dir_name = os.path.dirname(path)
        file_name = os.path.basename(path)
        descriptor['path'] = os.path.join(dir_name, hash, file_name)

    def row_counter(self, resource, iterator):
        counter = 0
        for row in iterator:
            counter += 1
            yield row
        DumperBase.inc_attr(self.datapackage.descriptor, self.datapackage_rowcount, counter)
        DumperBase.inc_attr(resource.res.descriptor, self.resource_rowcount, counter)
        resource.res.commit()
        self.datapackage.commit()

    def process_resources(self, resources):
        self.initialize()

        resource: ResourceWrapper = None
        for resource in resources:
            ret = self.process_resource(
                        ResourceWrapper(
                            resource.res,
                            schema_validator(resource.res, resource,
                                             **self.schema_validator_options)
                        )
            )
            ret = self.row_counter(resource, ret)
            yield ret

        # Calculate datapackage hash
        if self.datapackage_hash:
            datapackage_hash = hashlib.md5(
                        json.dumps(self.datapackage.descriptor,
                                   indent=2 if self.pretty_descriptor else None,
                                   sort_keys=True,
                                   ensure_ascii=True).encode('ascii')
                    ).hexdigest()
            DumperBase.set_attr(self.datapackage.descriptor, self.datapackage_hash, datapackage_hash)

        self.handle_datapackage()
        self.finalize()

    def handle_datapackage(self):
        self.datapackage.commit()
        self.stats['count_of_rows'] = DumperBase.get_attr(self.datapackage.descriptor, self.datapackage_rowcount)
        self.stats['bytes'] = DumperBase.get_attr(self.datapackage.descriptor, self.datapackage_bytes)
        self.stats['hash'] = DumperBase.get_attr(self.datapackage.descriptor, self.datapackage_hash)
        self.stats['dataset_name'] = self.datapackage.descriptor.get('name')

    def initialize(self):
        pass

    def finalize(self):
        pass



import os
import warnings
import datetime

from datapackage import Package
from tabulator import Stream
from tableschema.schema import Schema
from .. import DataStreamProcessor
from ..base.exceptions import SourceLoadError
from ..base.schema_validator import schema_validator, ignore, drop, raise_exception, clear
from ..helpers.resource_matcher import ResourceMatcher

from .parsers import XMLParser, ExcelXMLParser, ExtendedSQLParser


class StringsGuesser():
    def cast(self, value):
        return [('string', 'default', 0)]


class TypesGuesser():
    def cast(self, value):
        jts_type = {
            str: 'string',
            int: 'integer',
            float: 'number',
            list: 'array',
            dict: 'object',
            tuple: 'array',
            bool: 'boolean',
            datetime.datetime: 'datetime',
            datetime.date: 'date',
        }.get(type(value))
        ret = [('any', 'default', 0)]
        if jts_type is not None:
            ret.append(('jts_type', 'default', 1))
        return ret


class load(DataStreamProcessor):

    INFER_STRINGS = 'strings'
    INFER_PYTHON_TYPES = 'pytypes'
    INFER_FULL = 'full'

    CAST_TO_STRINGS = 'strings'
    CAST_DO_NOTHING = 'nothing'
    CAST_WITH_SCHEMA = 'schema'

    ERRORS_IGNORE = ignore
    ERRORS_DROP = drop
    ERRORS_RAISE = raise_exception
    ERRORS_CLEAR = clear

    def __init__(self, load_source, name=None, resources=None, strip=True, limit_rows=None,
                 infer_strategy=None, cast_strategy=None,
                 override_schema=None, override_fields=None,
                 extract_missing_values=None,
                 deduplicate_headers=False,
                 on_error=raise_exception,
                 **options):
        super(load, self).__init__()
        self.load_source = load_source

        self.name = name
        self.strip = strip
        self.limit_rows = limit_rows
        self.options = options
        self.resources = resources
        self.override_schema = override_schema
        self.override_fields = override_fields
        self.deduplicate_headers = deduplicate_headers

        # Extract missing values
        self.extract_missing_values = None
        if extract_missing_values is not None:
            if isinstance(extract_missing_values, bool):
                extract_missing_values = {}
            extract_missing_values.setdefault('source', None)
            extract_missing_values.setdefault('target', 'missingValues')
            extract_missing_values.setdefault('values', [])
            if isinstance(extract_missing_values.get('source'), str):
                extract_missing_values['source'] = [extract_missing_values['source']]
            self.extract_missing_values = extract_missing_values

        self.load_dp = None
        self.resource_descriptors = []
        self.iterators = []

        if 'force_strings' in options:
            warnings.warn('force_strings is being deprecated, use infer_strategy & cast_strategy instead',
                          DeprecationWarning)
            if options['force_strings']:
                infer_strategy = self.INFER_STRINGS
                cast_strategy = self.CAST_TO_STRINGS

        if 'validate' in options:
            warnings.warn('validate is being deprecated, use cast_strategy & on_error instead',
                          DeprecationWarning)
            if options['validate']:
                cast_strategy = self.CAST_WITH_SCHEMA

        # Force strings from stream for the INFER_STRINGS strategy
        if infer_strategy == self.INFER_STRINGS:
            self.options['force_strings'] = True

        self.guesser = {
            self.INFER_FULL: None,
            self.INFER_PYTHON_TYPES: TypesGuesser,
            self.INFER_STRINGS: StringsGuesser,
        }[infer_strategy or self.INFER_FULL]

        self.caster = {
            self.CAST_DO_NOTHING: lambda res, it: it,
            self.CAST_WITH_SCHEMA: lambda res, it: schema_validator(res, it, on_error=on_error),
            self.CAST_TO_STRINGS: lambda res, it: self.stringer(it)
        }[cast_strategy or self.CAST_DO_NOTHING]

    def process_datapackage(self, dp: Package):
        try:
            return self.safe_process_datapackage(dp)
        except Exception as e:
            raise SourceLoadError('Failed to load source {!r} and options {!r}: {}'
                                  .format(self.load_source, self.options, e)) from e

    def safe_process_datapackage(self, dp: Package):

        # If loading from datapackage & resource iterator:
        if isinstance(self.load_source, tuple):
            datapackage_descriptor, resource_iterator = self.load_source
            resources = datapackage_descriptor['resources']
            resource_matcher = ResourceMatcher(self.resources, datapackage_descriptor)
            for resource_descriptor in datapackage_descriptor['resources']:
                if resource_matcher.match(resource_descriptor['name']):
                    self.resource_descriptors.append(resource_descriptor)
            self.iterators = (resource for resource, descriptor in zip(resource_iterator, resources)
                              if resource_matcher.match(descriptor['name']))

        # If load_source is string:
        else:
            # Handle Environment vars if necessary:
            if self.load_source.startswith('env://'):
                env_var = self.load_source[6:]
                self.load_source = os.environ.get(env_var)
                if self.load_source is None:
                    raise ValueError(f"Couldn't find value for env var '{env_var}'")

            # Loading from datapackage:
            if os.path.basename(self.load_source) == 'datapackage.json' or self.options.get('format') == 'datapackage':
                self.load_dp = Package(self.load_source)
                resource_matcher = ResourceMatcher(self.resources, self.load_dp)
                for resource in self.load_dp.resources:
                    if resource_matcher.match(resource.name):
                        self.resource_descriptors.append(resource.descriptor)
                        self.iterators.append(resource.iter(keyed=True, cast=True))

            # Loading for any other source
            else:
                path = os.path.basename(self.load_source)
                path = os.path.splitext(path)[0]
                descriptor = dict(path=self.name or path,
                                  profile='tabular-data-resource')
                self.resource_descriptors.append(descriptor)
                descriptor['name'] = self.name or path
                if 'encoding' in self.options:
                    descriptor['encoding'] = self.options['encoding']
                self.options.setdefault('custom_parsers', {}).setdefault('xml', XMLParser)
                self.options.setdefault('custom_parsers', {}).setdefault('excel-xml', ExcelXMLParser)
                self.options.setdefault('custom_parsers', {}).setdefault('sql', ExtendedSQLParser)
                self.options.setdefault('ignore_blank_headers', True)
                self.options.setdefault('headers', 1)
                stream: Stream = Stream(self.load_source, **self.options).open()
                if len(stream.headers) != len(set(stream.headers)):
                    if not self.deduplicate_headers:
                        raise ValueError(
                            'Found duplicate headers.' +
                            'Use the `deduplicate_headers` flag (found headers=%r)' % stream.headers)
                    stream.headers = self.rename_duplicate_headers(stream.headers)
                schema = Schema().infer(
                    stream.sample, headers=stream.headers,
                    confidence=1, guesser_cls=self.guesser)
                # restore schema field names to original headers
                for header, field in zip(stream.headers, schema['fields']):
                    field['name'] = header
                if self.override_schema:
                    schema.update(self.override_schema)
                if self.override_fields:
                    fields = schema.get('fields', [])
                    for field in fields:
                        field.update(self.override_fields.get(field['name'], {}))
                if self.extract_missing_values:
                    missing_values = schema.get('missingValues', [])
                    if not self.extract_missing_values['values']:
                        self.extract_missing_values['values'] = missing_values
                    schema['fields'].append({
                        'name': self.extract_missing_values['target'],
                        'type': 'object',
                        'format': 'default',
                        'values': self.extract_missing_values['values'],
                    })
                descriptor['schema'] = schema
                descriptor['format'] = self.options.get('format', stream.format)
                descriptor['path'] += '.{}'.format(stream.format)
                self.iterators.append(stream.iter(keyed=True))
        dp.descriptor.setdefault('resources', []).extend(self.resource_descriptors)
        return dp

    def stripper(self, iterator):
        for r in iterator:
            yield dict(
                (k, v.strip()) if isinstance(v, str) else (k, v)
                for k, v in r.items()
            )

    def limiter(self, iterator):
        count = 0
        for row in iterator:
            yield row
            count += 1
            if count >= self.limit_rows:
                break

    def stringer(self, iterator):
        for r in iterator:
            yield dict(
                (k, str(v)) if not isinstance(v, str) else (k, v)
                for k, v in r.items()
            )

    def missing_values_extractor(self, iterator):
        source = self.extract_missing_values['source']
        target = self.extract_missing_values['target']
        values = self.extract_missing_values['values']
        for row in iterator:
            mapping = {}
            if values:
                for key, value in row.items():
                    if not source or key in source:
                        if value in values:
                            mapping[key] = value
            row[target] = mapping
            yield row

    def process_resources(self, resources):
        yield from super(load, self).process_resources(resources)
        for descriptor, it in zip(self.resource_descriptors, self.iterators):
            if self.extract_missing_values:
                it = self.missing_values_extractor(it)
            it = self.caster(descriptor, it)
            if self.strip:
                it = self.stripper(it)
            if self.limit_rows:
                it = self.limiter(it)
            yield it

    @staticmethod
    def rename_duplicate_headers(duplicate_headers):
        counter = {}
        headers = []
        for header in duplicate_headers:
            counter.setdefault(header, 0)
            counter[header] += 1
            if counter[header] > 1:
                if counter[header] == 2:
                    headers[headers.index(header)] = '%s (%s)' % (header, 1)
                header = '%s (%s)' % (header, counter[header])
            headers.append(header)
        return headers



from inspect import isfunction, signature
from collections.abc import Iterable

from .datastream_processor import DataStreamProcessor


class Flow:
    def __init__(self, *args):
        self.chain = args

    def results(self, on_error=None):
        return self._chain().results(on_error=on_error)

    def process(self):
        return self._chain().process()

    def datastream(self, ds=None):
        return self._chain(ds)._process()

    def _preprocess_chain(self):
        checkpoint_links = []
        for link in self.chain:
            if hasattr(link, 'handle_flow_checkpoint'):
                checkpoint_links = link.handle_flow_checkpoint(checkpoint_links)
            else:
                checkpoint_links.append(link)
        return checkpoint_links

    def _chain(self, ds=None):
        from ..helpers import datapackage_processor, rows_processor, row_processor, iterable_loader

        for position, link in enumerate(self._preprocess_chain(), start=1):
            if isinstance(link, Flow):
                ds = link._chain(ds)
            elif isinstance(link, DataStreamProcessor):
                ds = link(ds, position=position)
            elif isfunction(link):
                sig = signature(link)
                params = list(sig.parameters)
                if len(params) == 1:
                    if params[0] == 'row':
                        ds = row_processor(link)(ds, position=position)
                    elif params[0] == 'rows':
                        ds = rows_processor(link)(ds, position=position)
                    elif params[0] == 'package':
                        ds = datapackage_processor(link)(ds, position=position)
                    else:
                        assert False, 'Failed to parse function signature %r' % params
                else:
                    assert False, 'Failed to parse function signature %r' % params
            elif isinstance(link, Iterable):
                ds = iterable_loader(link)(ds, position=position)

        return ds



import copy
from kvfile import KVFile


def saver(resource, db, batch_size):
    gen = db.insert_generator(
        (('{:08x}'.format(idx), row)
         for idx, row
         in enumerate(resource)),
        batch_size=batch_size
    )
    for _, row in gen:
        yield row


def loader(db):
    for _, value in db.items():
        yield value


def duplicate(
    source=None,
    target_name=None,
    target_path=None,
    batch_size=1000,
    duplicate_to_end=False,
):
    def func(package):
        source_, target_name_, target_path_ = source, target_name, target_path
        if source_ is None:
            source_ = package.pkg.descriptor['resources'][0]['name']
        if target_name_ is None:
            target_name_ = source_ + '_copy'
        if target_path is None:
            target_path_ = target_name_ + '.csv'

        def traverse_resources(resources):
            new_res_list = []
            for res in resources:
                yield res
                if res['name'] == source_:
                    res = copy.deepcopy(res)
                    res['name'] = target_name_
                    res['path'] = target_path_
                    if duplicate_to_end:
                        new_res_list.append(res)
                    else:
                        yield res
            for res in new_res_list:
                yield res

        descriptor = package.pkg.descriptor
        descriptor['resources'] = list(traverse_resources(descriptor['resources']))
        yield package.pkg

        dbs = []
        for resource in package:
            if resource.res.name == source_:
                db = KVFile()
                yield saver(resource, db, batch_size)
                if duplicate_to_end:
                    dbs.append(db)
                else:
                    yield loader(db)
            else:
                yield resource
        for db in dbs:
            yield loader(db)

    return func



from ..helpers.resource_matcher import ResourceMatcher


def old_style_conditions(equals, not_equals):
    def func(row):
        return any(
                    (row[k] == v)
                    for o in equals
                    for k, v in o.items()
                ) or any(
                    (row[k] != v)
                    for o in not_equals
                    for k, v in o.items()
                )
    return func


def process_resource(rows, condition):
    for row in rows:
        if condition(row):
            yield row


def filter_rows(condition=None, equals=tuple(), not_equals=tuple(), resources=None):

    if not condition:
        condition = old_style_conditions(equals, not_equals)

    def func(package):
        matcher = ResourceMatcher(resources, package.pkg)
        yield package.pkg
        for r in package:
            if matcher.match(r.res.name):
                yield process_resource(r, condition)
            else:
                yield r

    return func



from .base import DataStream, DataStreamProcessor, schema_validator, ValidationError
from .base import ResourceWrapper, PackageWrapper
from .base import exceptions
from .base import Flow
from .processors import *  # noqa



from pathlib import Path
from .format_json import JSONFormat


class GeoJSONFormat(JSONFormat):

    def initialize_file(self, file):
        file.write('{"type": "FeatureCollection","features":')
        super(GeoJSONFormat, self).initialize_file(file)

    @classmethod
    def prepare_resource(cls, resource):
        super(GeoJSONFormat, cls).prepare_resource(resource)
        descriptor = resource.descriptor
        descriptor['path'] = str(Path(descriptor['path']).with_suffix('.geojson'))
        descriptor['format'] = 'geojson'
        descriptor['mediatype'] = 'application/geo+json'

    def write_transformed_row(self, transformed_row):
        properties = dict()
        geometry = None
        for k, v in transformed_row.items():
            if self.fields[k].type == 'geopoint':
                geometry = dict(
                    type='Point',
                    coordinates=v
                )
            elif self.fields[k].type == 'geojson':
                geometry = v
            else:
                properties[k] = v
        feature = dict(
            geometry=geometry,
            type='Feature',
            properties=properties
        )
        super(GeoJSONFormat, self).write_transformed_row(feature)

    def finalize_file(self):
        super(GeoJSONFormat, self).finalize_file()
        self.writer.write('}')



class DataflowsException(Exception):
    pass


class ProcessorError(DataflowsException):

    def __init__(self, cause, *, processor_name, processor_object, processor_position):
        self.cause = cause
        self.processor_name = processor_name
        self.processor_object = processor_object
        self.processor_position = processor_position
        super().__init__(str(cause))

    def __str__(self):
        return 'Errored in processor %s in position #%s: %s' % \
                    (self.processor_name, self.processor_position, self.cause)


class SourceLoadError(DataflowsException):
    pass



from dataflows import Flow, load, dump_to_path, dump_to_zip, printer, add_metadata
from dataflows import sort_rows, filter_rows, find_replace, delete_fields, set_type, validate, unpivot


{% if 'custom' in processing %}
def my_custom_processing(row):
    # Do some modifications to the row here:
    # ...
    return row
{% endif %}


def {{slug}}():
    flow = Flow(
        # Load inputs
        {% if input == 'file' %}
        load('{{input_url}}', format='{{format}}', {% if sheet %}sheet={{sheet}}{% endif %}),
        {% endif %}
        {% if input == 'remote' %}
        load('{{input_url}}', format='{{format}}', {% if sheet %}sheet={{sheet}}{% endif %}),
        {% endif %}
        {% if input == 'sql' %}
        load('{{input_url}}', table='{{input_db_table}}'),
        {% endif %}
        {% if input == 'other' %}
        {% endif %}
        # Process them (if necessary)
        {% if 'sort' in processing %}
        sort_rows('{field_name}'),  # Key is a Python format string or a list of field names
        {% endif %}
        {% if 'filter' in processing %}
        filter_rows(),
        {% endif %}
        {% if 'find_replace' in processing %}
        find_replace([
            dict(name='field_name',
                 patterns=[
                     dict(find='re-pattern-to-find', replace='re-pattern-to-replace-with'),                     
                 ])
        ]),
        {% endif %}
        {% if 'delete_fields' in processing %}
        delete_fields(['field_name']),  # Pass a list of field names to delete from the data
        {% endif %}
        {% if 'set_type' in processing %}
        set_type('field_name', type='number', constraints=dict(minimum=3)),  # There are quite a few options you can use here
                                                                             # Take a look at https://frictionlessdata.io/specs/table-schema/
        # Or you can simply use validate() here instead                                                                             
        {% endif %}
        {% if 'unpivot' in processing %}
        unpivot(unpivot_fields, extra_keys, extra_value),  # See documentation on the meaning of each of these parameters
        {% endif %}
        {% if 'custom' in processing %}
        my_custom_processing,
        {% endif %}
        # Save the results
        add_metadata(name='{{slug}}', title='''{{title}}'''),
        {% if output in ('print', 'print_n_pkg')  %}
        printer(),
        {% endif %}
        {% if output == 'list' %}
        {% endif %}
        {% if output in ('dp_csv', 'print_n_pkg') %}
        dump_to_path('{{slug}}'),
        {% endif %}
        {% if output == 'dp_csv_zip' %}
        dump_to_zip('{{slug}}.zip'),
        {% endif %}
        {% if output == 'dp_json' %}
        dump_to_path('{{slug}}', force_format=True, format='json'),
        {% endif %}
        {% if output == 'dp_json_zip' %}
        dump_to_zip('{{slug}}.zip', force_format=True, format='json'),
        {% endif %}
        {% if output == 'sql' %}
        dump_to_sql('{{output_url}}', table='{{output_db_table}}')
        {% endif %}
    )
    {% if output != 'list' %}
    flow.process()
    {% endif %}
    {% if output == 'list' %}
    data, *_ = flow.results()
    {% endif %}


if __name__ == '__main__':
    {{slug}}()



import re
import decimal
from kvfile import KVFile
from bitstring import BitArray
from ..helpers.resource_matcher import ResourceMatcher


class KeyCalc(object):
    def __init__(self, key_spec):
        self.calculator = self.__calculator(key_spec)

    def __calculator(self, key_spec):
        if callable(key_spec):
            return key_spec
        if isinstance(key_spec, str):
            key_list = re.findall(r'\{(.*?)\}', key_spec)

            def func(row):
                context = row.copy()
                for key, value in row.items():
                    # We need to stringify some types to make them properly comparable
                    if key in key_list:
                        # numbers
                        # https://www.h-schmidt.net/FloatConverter/IEEE754.html
                        if isinstance(value, (int, float, decimal.Decimal)):
                            bits = BitArray(float=value, length=64)
                            # invert the sign bit
                            bits.invert(0)
                            # invert negative numbers
                            if value < 0:
                                bits.invert(range(1, 64))
                            context[key] = bits.hex
                return key_spec.format(**context)
            return func
        assert False, 'key should be either a format string or a row->string callable'

    def __call__(self, row):
        return self.calculator(row)


def _sorter(rows, key_calc, reverse, batch_size):
    db = KVFile()

    def process(rows):
        for row_num, row in enumerate(rows):
            key = key_calc(row) + '{:08x}'.format(row_num)
            yield (key, row)

    db.insert(process(rows), batch_size=batch_size)
    for _, value in db.items(reverse=reverse):
        yield value


def sort_rows(key, resources=None, reverse=False, batch_size=1000):
    key_calc = KeyCalc(key)

    def func(package):
        matcher = ResourceMatcher(resources, package.pkg)
        yield package.pkg
        for rows in package:
            if matcher.match(rows.res.name):
                yield _sorter(rows, key_calc, reverse, batch_size)
            else:
                yield rows

    return func



from datapackage import Package


class PackageWrapper:
    def __init__(self, pkg: Package):
        self.pkg: Package = pkg
        self.it = iter([])
        assert isinstance(pkg, Package)

    def __iter__(self):
        return self.it
